68b542363f
Add scalable init API * Add new ncclCommInitRankScalable to allow for passing multiple unique IDs to the init function. * Spreads the load onto multiple bootstrap roots, allowing for constant bootstrap time. * Requires multiple ranks to create a unique ID, and the CPU-side ID exchange code to call allgather[v] instead of broadcast. Accelerate init bootstrap operations * Reduce the number of calls to allgather. * Allow roots to reply early to ranks when information is already available. * Add an option to use ncclNet instead of sockets to perform bootstrap allgather operations. Add PAT algorithms for Allgather and ReduceScatter * Parallel Aggregated Trees, variation of Bruck algorithm. * Logarithmic number of network steps for small sizes at scale. * Only supports one rank per node at the moment. Add support for registered buffers for intra-node communication. * Allow registered user buffers to be accessed directly intra-node * Avoids extra copies in algorithms which permit it, saving memory bandwidth and helping with compute overlap. Add profiler plugin API * New plugin API for profiling * Supports various levels of profiling, with a hierarchy. Asynchronous graph allocation * Make calls to cudaMalloc and cudaMemcpy during graph allocation asynchronous. * Significantly speeds up graph capture. Use fatal IB asynchronous events to stop network operation * Avoids many other error messages * Only fatal errors are affected; potentially transient errors (e.g. port down) do not cause an immediate stop. Set P2P level to PXB on AMD CPUs when using more than 2 GPUs per node * P2P would cause a significant performance degradation when using many GPUs, and therefore many interleaved data flows. * Disable P2P through the CPU when we have 3+ GPUs per node; keep it enabled when we only have 2 GPUs. Improve the init logs to report the real NCCL function. * Make the log report ncclCommInitRank or ncclCommSplit, rather than the generic ncclCommInitRankFunc. Add a parameter to set the location of the user configuration file. * Add NCCL_CONF_FILE environment variable to set where the user's configuration file resides. Increase default IB timeout * Increase IB timeout value from 18 to 20. * Should help avoid fatal errors on large RoCE systems. Add new check for nvidia peermem * On linux kernels 6.6+, /sys/kernel/mm/memory_peers is no longer present; check for /sys/module/nvidia_peermem/version instead. Fix old performance regression when mixing small and large operations. * Improves distribution of work on channels. Fix crash when NUMA IDs are equal to -1. * Can happen when a NIC is a virtual NIC, or when linux doesn't know which NUMA node a device is attached to * Issue NVIDIA/nccl-tests#233 Fix tree graph search when NCCL_CROSS_NIC is set to 1. * Would force NCCL to use the balanced_tree pattern, thereby disabling LL128 on platforms with 1 GPU+1 NIC per PCI switch. * Would also try to use alternate rings even though it was not needed. Compiler tweaks and fixes * PR #1177 * PR #1228 Fix stack smash * PR #1325 Fixes for multi-node NVLink + IB operation Coverity fixes and comments.
217 řádky
7.6 KiB
C++
217 řádky
7.6 KiB
C++
/*************************************************************************
|
|
* Copyright (c) 2016-2022, NVIDIA CORPORATION. All rights reserved.
|
|
*
|
|
* See LICENSE.txt for license information
|
|
************************************************************************/
|
|
|
|
#include "shmutils.h"
|
|
#include "comm.h"
|
|
#include "checks.h"
|
|
#include <sys/types.h>
|
|
#include <sys/mman.h>
|
|
#include <sys/stat.h>
|
|
#include <fcntl.h>
|
|
#include <errno.h>
|
|
#include <stdio.h>
|
|
#include <stdlib.h>
|
|
#include <unistd.h>
|
|
#include <utils.h>
|
|
|
|
struct shmHandleInternal {
|
|
int fd;
|
|
char* shmPath;
|
|
char* shmPtr;
|
|
void* devShmPtr;
|
|
size_t shmSize;
|
|
size_t realShmSize;
|
|
int* refcount;
|
|
};
|
|
|
|
static void shmHandleInit(int fd, char* shmPath, size_t shmSize, size_t realShmSize, char* hptr, void* dptr, bool create, struct shmHandleInternal* handle) {
|
|
handle->fd = fd;
|
|
handle->shmPtr = hptr;
|
|
handle->devShmPtr = dptr;
|
|
handle->shmSize = shmSize;
|
|
handle->realShmSize = realShmSize;
|
|
handle->refcount = (hptr != NULL) ? (int*)(hptr + shmSize) : NULL;
|
|
if (create) {
|
|
int slen = strlen(shmPath);
|
|
handle->shmPath = (char*)malloc(slen + 1);
|
|
memcpy(handle->shmPath, shmPath, slen + 1);
|
|
if (hptr) memset(hptr, 0, shmSize);
|
|
} else {
|
|
handle->shmPath = NULL;
|
|
}
|
|
return;
|
|
}
|
|
|
|
ncclResult_t ncclShmOpen(char* shmPath, size_t shmSize, void** shmPtr, void** devShmPtr, int refcount, ncclShmHandle_t* handle) {
|
|
int fd = -1;
|
|
char* hptr = NULL;
|
|
void* dptr = NULL;
|
|
ncclResult_t ret = ncclSuccess;
|
|
struct shmHandleInternal* tmphandle;
|
|
bool create = refcount > 0 ? true : false;
|
|
const size_t refSize = sizeof(int); /* extra sizeof(int) bytes for reference count */
|
|
const size_t realShmSize = shmSize + refSize;
|
|
|
|
*handle = *shmPtr = NULL; /* assume shmPtr and handle always set correctly by users. */
|
|
EQCHECKGOTO(tmphandle = (struct shmHandleInternal*)calloc(1, sizeof(struct shmHandleInternal)), NULL, ret, fail);
|
|
if (create) {
|
|
/* refcount > 0 means the caller tries to allocate a shared memory. This shared memory segment will have
|
|
* refcount references; when the peer attaches, it should pass -1 to reduce one reference count. When it
|
|
* goes down to 0, unlink should be called in order to delete shared memory file. */
|
|
if (shmPath[0] == '\0') {
|
|
sprintf(shmPath, "/dev/shm/nccl-XXXXXX");
|
|
retry_mkstemp:
|
|
fd = mkstemp(shmPath);
|
|
if (fd < 0) {
|
|
if (errno == EINTR) {
|
|
INFO(NCCL_ALL, "mkstemp: Failed to create %s, error: %s (%d) - retrying", shmPath, strerror(errno), errno);
|
|
goto retry_mkstemp;
|
|
}
|
|
WARN("Error: failed to create shared memory file %p, error %s (%d)", shmPath, strerror(errno), errno);
|
|
ret = ncclSystemError;
|
|
goto fail;
|
|
}
|
|
} else {
|
|
SYSCHECKGOTO(fd = open(shmPath, O_CREAT | O_RDWR, S_IRUSR | S_IWUSR), "open", ret, fail);
|
|
}
|
|
|
|
retry_fallocate:
|
|
if (fallocate(fd, 0, 0, realShmSize) != 0) {
|
|
if (errno == EINTR) {
|
|
INFO(NCCL_ALL, "fallocate: Failed to extend %s to %ld bytes, error: %s (%d) - retrying", shmPath, realShmSize, strerror(errno), errno);
|
|
goto retry_fallocate;
|
|
}
|
|
WARN("Error: failed to extend %s to %ld bytes, error: %s (%d)", shmPath, realShmSize, strerror(errno), errno);
|
|
ret = ncclSystemError;
|
|
goto fail;
|
|
}
|
|
INFO(NCCL_ALLOC, "Allocated %ld bytes of shared memory in %s", realShmSize, shmPath);
|
|
} else {
|
|
SYSCHECKGOTO(fd = open(shmPath, O_RDWR, S_IRUSR | S_IWUSR), "open", ret, fail);
|
|
}
|
|
|
|
hptr = (char*)mmap(NULL, realShmSize, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
|
|
if (hptr == MAP_FAILED) {
|
|
WARN("Error: Could not map %s size %zu, error: %s (%d)", shmPath, realShmSize, strerror(errno), errno);
|
|
ret = ncclSystemError;
|
|
hptr = NULL;
|
|
goto fail;
|
|
}
|
|
|
|
if (create) {
|
|
*(int*)(hptr + shmSize) = refcount;
|
|
} else {
|
|
int remref = ncclAtomicRefCountDecrement((int*)(hptr + shmSize));
|
|
if (remref == 0) {
|
|
/* the last peer has completed attachment, it should unlink the shm mem file. */
|
|
if (unlink(shmPath) != 0) {
|
|
INFO(NCCL_ALLOC, "unlink shared memory %s failed, error: %s (%d)", shmPath, strerror(errno), errno);
|
|
}
|
|
}
|
|
}
|
|
|
|
if (devShmPtr) {
|
|
CUDACHECKGOTO(cudaHostRegister((void*)hptr, realShmSize, cudaHostRegisterPortable | cudaHostRegisterMapped), ret, fail);
|
|
CUDACHECKGOTO(cudaHostGetDevicePointer(&dptr, (void*)hptr, 0), ret, fail);
|
|
}
|
|
|
|
shmHandleInit(fd, shmPath, shmSize, realShmSize, hptr, dptr, create, tmphandle);
|
|
exit:
|
|
*shmPtr = hptr;
|
|
if (devShmPtr) *devShmPtr = dptr;
|
|
*handle = (ncclShmHandle_t)tmphandle;
|
|
return ret;
|
|
fail:
|
|
WARN("Error while %s shared memory segment %s (size %ld), error: %s (%d)", create ? "creating" : "attaching to",
|
|
shmPath, shmSize, strerror(errno), errno);
|
|
if (tmphandle) {
|
|
shmHandleInit(fd, shmPath, shmSize, realShmSize, hptr, dptr, create, tmphandle);
|
|
(void)ncclShmClose((ncclShmHandle_t)tmphandle);
|
|
tmphandle = NULL;
|
|
}
|
|
hptr = NULL;
|
|
dptr = NULL;
|
|
goto exit;
|
|
}
|
|
|
|
ncclResult_t ncclShmClose(ncclShmHandle_t handle) {
|
|
ncclResult_t ret = ncclSuccess;
|
|
struct shmHandleInternal* tmphandle = (struct shmHandleInternal*)handle;
|
|
if (tmphandle) {
|
|
if (tmphandle->fd >= 0) {
|
|
close(tmphandle->fd);
|
|
if (tmphandle->shmPath != NULL && tmphandle->refcount != NULL && *tmphandle->refcount > 0) {
|
|
if (unlink(tmphandle->shmPath) != 0) {
|
|
WARN("unlink shared memory %s failed, error: %s (%d)", tmphandle->shmPath, strerror(errno), errno);
|
|
ret = ncclSystemError;
|
|
}
|
|
}
|
|
free(tmphandle->shmPath);
|
|
}
|
|
|
|
if (tmphandle->shmPtr) {
|
|
if (tmphandle->devShmPtr) CUDACHECK(cudaHostUnregister(tmphandle->shmPtr));
|
|
if (munmap(tmphandle->shmPtr, tmphandle->realShmSize) != 0) {
|
|
WARN("munmap of shared memory %p size %ld failed, error: %s (%d)", tmphandle->shmPtr, tmphandle->realShmSize, strerror(errno), errno);
|
|
ret = ncclSystemError;
|
|
}
|
|
}
|
|
free(tmphandle);
|
|
}
|
|
return ret;
|
|
}
|
|
|
|
ncclResult_t ncclShmUnlink(ncclShmHandle_t handle) {
|
|
ncclResult_t ret = ncclSuccess;
|
|
struct shmHandleInternal* tmphandle = (struct shmHandleInternal*)handle;
|
|
if (tmphandle) {
|
|
if (tmphandle->shmPath != NULL && tmphandle->refcount != NULL && *tmphandle->refcount > 0) {
|
|
if (unlink(tmphandle->shmPath) != 0) {
|
|
WARN("unlink shared memory %s failed, error: %s (%d)", tmphandle->shmPath, strerror(errno), errno);
|
|
ret = ncclSystemError;
|
|
}
|
|
free(tmphandle->shmPath);
|
|
tmphandle->shmPath = NULL;
|
|
}
|
|
}
|
|
return ret;
|
|
}
|
|
|
|
ncclResult_t ncclShmemAllgather(struct ncclComm *comm, struct ncclShmemCollBuff *shmem, void *sendbuff, void *recvbuff, size_t typeSize) {
|
|
ncclResult_t ret = ncclSuccess;
|
|
int curRound;
|
|
size_t mycnt;
|
|
|
|
if (comm == NULL || shmem == NULL || sendbuff == NULL || recvbuff == NULL || shmem->maxTypeSize < typeSize) {
|
|
ret = ncclInvalidArgument;
|
|
goto exit;
|
|
}
|
|
|
|
curRound = shmem->round;
|
|
memcpy((char*)shmem->ptr[curRound] + comm->localRank * typeSize, sendbuff, typeSize);
|
|
/* sync among local ranks */
|
|
mycnt = __atomic_add_fetch(shmem->cnt[curRound], 1, __ATOMIC_ACQ_REL);
|
|
if (mycnt == comm->localRanks) {
|
|
*shmem->cnt[curRound ^ 1] = 0; /* prepare next round */
|
|
__atomic_store_n(shmem->cnt[curRound], comm->localRanks + 1, __ATOMIC_RELEASE); /* release everyone */
|
|
} else {
|
|
uint64_t t0 = clockNano();
|
|
while(__atomic_load_n(shmem->cnt[curRound], __ATOMIC_ACQUIRE) != comm->localRanks + 1) {
|
|
if (clockNano() - t0 >= 5 * 1000) sched_yield();
|
|
if (__atomic_load_n(comm->abortFlag, __ATOMIC_ACQUIRE) == 1) {
|
|
ret = ncclInternalError;
|
|
goto exit;
|
|
}
|
|
}
|
|
}
|
|
|
|
memcpy(recvbuff, (const void*)shmem->ptr[curRound], comm->localRanks * typeSize);
|
|
shmem->round ^= 1;
|
|
|
|
exit:
|
|
return ret;
|
|
}
|